@@ -1,5 +1,5 @@
 /*****************************************************************************
- *  $Id: gtkcompletionline.cc,v 1.24 2001/10/19 09:19:35 mishoo Exp $
+ *  $Id: gtkcompletionline.cc,v 1.31.2.1 2003/06/21 23:14:34 mishoo Exp $
  *  Copyright (C) 2000, Mishoo
  *  Author: Mihai Bazon                  Email: mishoo@fenrir.infoiasi.ro
  *
@@ -10,12 +10,12 @@
  *****************************************************************************/
 
 
-#include "gtkcompletionline.h"
 #include <gdk/gdkkeysyms.h>
 #include <gtk/gtksignal.h>
 #include <gtk/gtkclist.h>
 #include <gtk/gtkwindow.h>
 #include <gtk/gtkscrolledwindow.h>
+#include <gtk/gtkmain.h>
 
 #include <stddef.h>
 #include <stdio.h>
@@ -25,12 +25,15 @@
 #include <sys/stat.h>
 #include <unistd.h>
 
+#include <iostream>
 #include <set>
-#include <strstream>
+#include <sstream>
 #include <string>
 #include <vector>
 using namespace std;
 
+#include "gtkcompletionline.h"
+
 static int on_row_selected_handler = 0;
 static int on_key_press_handler = 0;
 
@@ -45,6 +48,8 @@ enum {
   SEARCH_MODE,
   SEARCH_LETTER,
   SEARCH_NOT_FOUND,
+  EXT_HANDLER,
+  CANCEL,
   LAST_SIGNAL
 };
 
@@ -61,6 +66,7 @@ static StrSet path;
 static StrSet execs;
 static StrSet dirlist;
 static string prefix;
+static int g_show_dot_files;
 
 /* callbacks */
 static void gtk_completion_line_class_init(GtkCompletionLineClass *klass);
@@ -100,55 +106,71 @@ gtk_completion_line_class_init(GtkCompletionLineClass *klass)
 
   gtk_completion_line_signals[UNIQUE] =
     gtk_signal_new("unique",
-                   GTK_RUN_FIRST, object_class->type,
+                   GTK_RUN_FIRST, G_TYPE_FROM_CLASS(object_class),
                    GTK_SIGNAL_OFFSET(GtkCompletionLineClass,
                                      unique),
                    gtk_signal_default_marshaller, GTK_TYPE_NONE, 0);
 
   gtk_completion_line_signals[NOTUNIQUE] =
     gtk_signal_new("notunique",
-                   GTK_RUN_FIRST, object_class->type,
+                   GTK_RUN_FIRST, G_TYPE_FROM_CLASS(object_class),
                    GTK_SIGNAL_OFFSET(GtkCompletionLineClass,
                                      notunique),
                    gtk_signal_default_marshaller, GTK_TYPE_NONE, 0);
 
   gtk_completion_line_signals[INCOMPLETE] =
     gtk_signal_new("incomplete",
-                   GTK_RUN_FIRST, object_class->type,
+                   GTK_RUN_FIRST, G_TYPE_FROM_CLASS(object_class),
                    GTK_SIGNAL_OFFSET(GtkCompletionLineClass,
                                      incomplete),
                    gtk_signal_default_marshaller, GTK_TYPE_NONE, 0);
 
   gtk_completion_line_signals[RUNWITHTERM] =
     gtk_signal_new("runwithterm",
-                   GTK_RUN_FIRST, object_class->type,
+                   GTK_RUN_FIRST, G_TYPE_FROM_CLASS(object_class),
                    GTK_SIGNAL_OFFSET(GtkCompletionLineClass,
                                      runwithterm),
                    gtk_signal_default_marshaller, GTK_TYPE_NONE, 0);
 
   gtk_completion_line_signals[SEARCH_MODE] =
     gtk_signal_new("search_mode",
-                   GTK_RUN_FIRST, object_class->type,
+                   GTK_RUN_FIRST, G_TYPE_FROM_CLASS(object_class),
                    GTK_SIGNAL_OFFSET(GtkCompletionLineClass,
                                      search_mode),
                    gtk_signal_default_marshaller, GTK_TYPE_NONE, 0);
 
   gtk_completion_line_signals[SEARCH_NOT_FOUND] =
     gtk_signal_new("search_not_found",
-                   GTK_RUN_FIRST, object_class->type,
+                   GTK_RUN_FIRST, G_TYPE_FROM_CLASS(object_class),
                    GTK_SIGNAL_OFFSET(GtkCompletionLineClass,
                                      search_not_found),
                    gtk_signal_default_marshaller, GTK_TYPE_NONE, 0);
 
   gtk_completion_line_signals[SEARCH_LETTER] =
     gtk_signal_new("search_letter",
-                   GTK_RUN_FIRST, object_class->type,
+                   GTK_RUN_FIRST, G_TYPE_FROM_CLASS(object_class),
                    GTK_SIGNAL_OFFSET(GtkCompletionLineClass,
                                      search_letter),
                    gtk_signal_default_marshaller, GTK_TYPE_NONE, 0);
 
-  gtk_object_class_add_signals(object_class,
-                               gtk_completion_line_signals, LAST_SIGNAL);
+  gtk_completion_line_signals[EXT_HANDLER] =
+    gtk_signal_new("ext_handler",
+                   GTK_RUN_FIRST, G_TYPE_FROM_CLASS(object_class),
+                   GTK_SIGNAL_OFFSET(GtkCompletionLineClass,
+                                     ext_handler),
+                   gtk_marshal_NONE__POINTER, GTK_TYPE_NONE, 1,
+                   GTK_TYPE_POINTER);
+
+  gtk_completion_line_signals[CANCEL] =
+    gtk_signal_new("cancel",
+                   GTK_RUN_FIRST, G_TYPE_FROM_CLASS(object_class),
+                   GTK_SIGNAL_OFFSET(GtkCompletionLineClass,
+                                     ext_handler),
+                   gtk_marshal_NONE__POINTER, GTK_TYPE_NONE, 1,
+                   GTK_TYPE_POINTER);
+
+  //gtk_object_class_add_signals(object_class,
+  //                             gtk_completion_line_signals, LAST_SIGNAL);
 
   klass->unique = NULL;
   klass->notunique = NULL;
@@ -157,6 +179,8 @@ gtk_completion_line_class_init(GtkCompletionLineClass *klass)
   klass->search_mode = NULL;
   klass->search_letter = NULL;
   klass->search_not_found = NULL;
+  klass->ext_handler = NULL;
+  klass->cancel = NULL;
 }
 
 /* init */
@@ -171,6 +195,8 @@ gtk_completion_line_init(GtkCompletionLine *object)
   object->list_compl = NULL;
   object->hist_search_mode = GCL_SEARCH_OFF;
   object->hist_word = new string;
+  object->tabtimeout = 0;
+  object->show_dot_files = 0;
 
   on_key_press_handler =
     gtk_signal_connect(GTK_OBJECT(object), "key_press_event",
@@ -180,15 +206,62 @@ gtk_completion_line_init(GtkCompletionLine *object)
 
   object->hist = new HistoryFile();
 
+  object->first_key = 1;
+}
+
+void gtk_completion_line_last_history_item(GtkCompletionLine* object) {
   const char *last_item = object->hist->last_item();
   if (last_item) {
-    gtk_entry_set_text(GTK_ENTRY(object), last_item);
-    gtk_entry_select_region(GTK_ENTRY(object), 0, 
-                            strlen(object->hist->last_item()));
+    object->hist->set_default("");
+    const char* txt = object->hist->prev();
+    gtk_entry_set_text(GTK_ENTRY(object), txt);
+    gtk_entry_select_region(GTK_ENTRY(object), 0, strlen(txt));
   }
 }
 
-static int
+string quote_string(const string& str)
+{
+  string res;
+  const char* i = str.c_str();
+  while (*i) {
+    char c = *i++;
+    switch (c) {
+     case ' ':
+      res += '\\';
+     default:
+      res += c;
+    }
+  }
+  return res;
+}
+
+static void
+get_token(istream& is, string& s)
+{
+  s.clear();
+  bool escaped = false;
+  while (!is.eof()) {
+    char c = is.get();
+    if (is.eof())
+      break;
+    if (escaped) {
+      s += c;
+      escaped = false;
+    } else if (c == '\\') {
+      // s += c;
+      escaped = true;
+    } else if (::isspace(c)) {
+      while (::isspace(c) && !is.eof()) c = is.get();
+      if (!is.eof())
+        is.unget();
+      break;
+    } else {
+      s += c;
+    }
+  }
+}
+
+int
 get_words(GtkCompletionLine *object, vector<string>& words)
 {
   string content(gtk_entry_get_text(GTK_ENTRY(object)));
@@ -199,49 +272,55 @@ get_words(GtkCompletionLine *object, vector<string>& words)
     if (i != content.end())
       content.insert(i, ' ');
   }
-  istrstream ss(content.c_str());
+  istringstream ss(content);
 
   while (!ss.eof()) {
     string s;
-    ss >> s;
+    // ss >> s;
+    get_token(ss, s);
     words.push_back(s);
     if (ss.eof()) break;
-    if (ss.tellg() < pos_in_text && ss.tellg() >= 0) {
-      pos++;
-    }
+    if (ss.tellg() < pos_in_text && ss.tellg() >= 0)
+      ++pos;
   }
 
   return pos;
 }
 
-static int
+int
 set_words(GtkCompletionLine *object, const vector<string>& words, int pos = -1)
 {
-  strstream ss;
-  bool first = true;
-  if (pos == -1) {
+  ostringstream ss;
+  if (pos == -1)
     pos = words.size() - 1;
-  }
   int where = 0;
 
-  for (vector<string>::const_iterator i = words.begin();
-       i != words.end(); i++) {
-    if (first) {
-      first = false;
-    } else {
-      ss << " ";
-    }
-    ss << *i;
-    if (pos == 0 && where == 0) {
+  vector<string>::const_iterator
+    i     = words.begin(),
+    i_end = words.end();
+
+  while (i != i_end) {
+    ss << quote_string(*i++);
+    if (i != i_end)
+      ss << ' ';
+    if (!pos && !where)
       where = ss.tellp();
-    } else {
-      pos--;
-    }
+    else
+      --pos;
   }
   ss << ends;
 
-  gtk_entry_set_text(GTK_ENTRY(object), ss.str());
-  delete[] ss.str();
+  if (words.size() == 1) {
+    const string& s = words.back();
+    size_t pos = s.rfind('.');
+    if (pos != string::npos)
+      gtk_signal_emit_by_name(
+        GTK_OBJECT(object), "ext_handler", s.substr(pos + 1).c_str());
+    else
+      gtk_signal_emit_by_name(GTK_OBJECT(object), "ext_handler", NULL);
+  }
+
+  gtk_entry_set_text(GTK_ENTRY(object), ss.str().c_str());
   gtk_editable_set_position(GTK_EDITABLE(object), where);
   return where;
 }
@@ -251,7 +330,7 @@ generate_path()
 {
   char *path_cstr = (char*)getenv("PATH");
 
-  istrstream path_ss(path_cstr);
+  istringstream path_ss(path_cstr);
   string tmp;
 
   path.clear();
@@ -263,7 +342,8 @@ generate_path()
       if (c == ':' || path_ss.eof()) break;
       else tmp += c;
     } while (true);
-    if (tmp.length() != 0) path.insert(tmp);
+    if (tmp.length() != 0)
+      path.insert(tmp);
   }
 }
 
@@ -272,21 +352,24 @@ select_executables_only(const struct dirent* dent)
 {
   int len = strlen(dent->d_name);
   int lenp = prefix.length();
-  struct stat stat_data;
 
   if (dent->d_name[0] == '.') {
-    if (dent->d_name[1] == '\0') return 0;
-    if (dent->d_name[1] == '.') {
-      if (dent->d_name[2] == '\0') return 0;
-    }
+    if (!g_show_dot_files)
+      return 0;
+    if (dent->d_name[1] == '\0')
+      return 0;
+    if ((dent->d_name[1] == '.') && (dent->d_name[2] == '\0'))
+      return 0;
   }
-  if (dent->d_name[len - 1] == '~') return 0;
-  if (lenp == 0) return 1;
-  if (lenp > len) return 0;
+  if (dent->d_name[len - 1] == '~')
+    return 0;
+  if (lenp == 0)
+    return 1;
+  if (lenp > len)
+    return 0;
 
-  if (strncmp(dent->d_name, prefix.c_str(), lenp) == 0) {
+  if (strncmp(dent->d_name, prefix.c_str(), lenp) == 0)
     return 1;
-  }
 
   return 0;
 }
@@ -297,7 +380,7 @@ int my_alphasort(const void* va, const void* vb) {
 
   const char* s1 = (*a)->d_name;
   const char* s2 = (*b)->d_name;
-  
+
   int l1 = strlen(s1);
   int l2 = strlen(s2);
   int result = strcmp(s1, s2);
@@ -349,13 +432,10 @@ generate_completion_from_execs(GtkCompletionLine *object)
 }
 
 static string
-get_common_part(const char *s1, const char *s2)
+get_common_part(const char *p1, const char *p2)
 {
   string ret;
 
-  const char *p1 = s1;
-  const char *p2 = s2;
-
   while (*p1 == *p2 && *p1 != '\0' && *p2 != '\0') {
     ret += *p1;
     p1++;
@@ -374,11 +454,11 @@ complete_common(GtkCompletionLine *object)
   int pos = get_words(object, words);
   words[pos] = ((GString*)ls->data)->str;
 
-  ls = ls->next;
+  ls = g_list_next(ls);
   while (ls != NULL) {
     words[pos] = get_common_part(words[pos].c_str(),
                                  ((GString*)ls->data)->str);
-    ls = ls->next;
+    ls = g_list_next(ls);
   }
 
   set_words(object, words, pos);
@@ -409,11 +489,12 @@ generate_dirlist(const char *what)
     dest += *p;
     if (*p == '/') {
       DIR* dir = opendir(dest.c_str());
-      if (!dir) goto dirty;
+      if (!dir)
+        goto dirty;
       closedir(dir);
       filename = p;
     }
-    p++;
+    ++p;
   }
 
   *filename = '\0';
@@ -442,6 +523,7 @@ generate_dirlist(const char *what)
 
   free(str);
   return GEN_COMPLETION_OK;
+
  dirty:
   free(str);
   return GEN_CANT_COMPLETE;
@@ -466,16 +548,19 @@ static int
 parse_tilda(GtkCompletionLine *object)
 {
   string text = gtk_entry_get_text(GTK_ENTRY(object));
-  int where = text.find("~");
+  gint where = (gint)text.find("~");
   if (where != string::npos) {
-    if (where > 0) {
-      if (text[where - 1] != ' ') return 0;
-    }
+    if ((where > 0) && (text[where - 1] != ' '))
+      return 0;
     if (where < text.size() - 1 && text[where + 1] != '/') {
-      // Parse user's home
+      // FIXME: Parse another user's home
     } else {
-      text.replace(where, 1, g_get_home_dir());
-      gtk_entry_set_text(GTK_ENTRY(object), text.c_str());
+      string home = g_get_home_dir();
+      size_t i = home.length() - 1;
+      while ((i >= 0) && (home[i] == '/'))
+        home.erase(i--);
+      gtk_editable_insert_text(GTK_EDITABLE(object), home.c_str(), home.length(), &where);
+      gtk_editable_delete_text(GTK_EDITABLE(object), where, where + 1);
     }
   }
 
@@ -494,12 +579,10 @@ complete_from_list(GtkCompletionLine *object)
   if (object->win_compl != NULL) {
     object->where = (GList*)gtk_clist_get_row_data(
       GTK_CLIST(object->list_compl), object->list_compl_items_where);
-
     words[pos] = ((GString*)object->where->data)->str;
     int current_pos = set_words(object, words, pos);
     gtk_entry_select_region(GTK_ENTRY(object),
                             object->pos_in_text, current_pos);
-
     int &item = object->list_compl_items_where;
     gtk_clist_select_row(GTK_CLIST(object->list_compl), item, 0);
     gtk_clist_moveto(GTK_CLIST(object->list_compl), item, 0, 0.5, 0.0);
@@ -509,12 +592,11 @@ complete_from_list(GtkCompletionLine *object)
     int current_pos = set_words(object, words, pos);
     gtk_entry_select_region(GTK_ENTRY(object),
                             object->pos_in_text, current_pos);
-    object->where = object->where->next;
-    g_list_next(object->where);
+    object->where = g_list_next(object->where);
   }
 }
 
-static int
+static void
 on_row_selected(GtkWidget *ls, gint row, gint col, GdkEvent *ev, gpointer data)
 {
   GtkCompletionLine *cl = GTK_COMPLETION_LINE(data);
@@ -527,6 +609,8 @@ on_row_selected(GtkWidget *ls, gint row, gint col, GdkEvent *ev, gpointer data)
                              on_row_selected_handler);
 }
 
+/*
+
 static void
 select_appropiate(GtkCompletionLine *object)
 {
@@ -554,6 +638,8 @@ get_prefix(GtkCompletionLine *object)
   prefix = words[pos];
 }
 
+*/
+
 static int
 complete_line(GtkCompletionLine *object)
 {
@@ -562,6 +648,7 @@ complete_line(GtkCompletionLine *object)
   int pos = get_words(object, words);
   prefix = words[pos];
 
+  g_show_dot_files = object->show_dot_files;
   if (prefix[0] != '/') {
     if (object->where == NULL) {
       generate_path();
@@ -569,12 +656,10 @@ complete_line(GtkCompletionLine *object)
       generate_completion_from_execs(object);
       object->where = NULL;
     }
-  } else {
-    if (object->where == NULL) {
-      generate_dirlist(prefix.c_str());
-      generate_completion_from_dirlist(object);
-      object->where = NULL;
-    }
+  } else if (object->where == NULL) {
+    generate_dirlist(prefix.c_str());
+    generate_completion_from_dirlist(object);
+    object->where = NULL;
   }
 
   if (object->cmpl != NULL) {
@@ -586,10 +671,9 @@ complete_line(GtkCompletionLine *object)
   if (object->where != NULL) {
     if (object->win_compl != NULL) {
       int &item = object->list_compl_items_where;
-      item++;
-      if (item >= object->list_compl_nr_rows) {
+      ++item;
+      if (item >= object->list_compl_nr_rows)
         item = object->list_compl_nr_rows - 1;
-      }
     }
     complete_from_list(object);
   } else if (object->cmpl != NULL) {
@@ -615,12 +699,13 @@ complete_line(GtkCompletionLine *object)
 
       if (object->win_compl == NULL) {
         object->win_compl = gtk_window_new(GTK_WINDOW_POPUP);
+        gtk_widget_set_name(object->win_compl, "Msh_Run_Window");
+
         /*gtk_window_set_position(GTK_WINDOW(object->win_compl),
           GTK_WIN_POS_MOUSE);*/
 
         gtk_window_set_policy(GTK_WINDOW(object->win_compl),
                               FALSE, FALSE, TRUE);
-        gtk_container_set_border_width(GTK_CONTAINER(object->win_compl), 5);
 
         object->list_compl = gtk_clist_new(1);
 
@@ -644,18 +729,19 @@ complete_line(GtkCompletionLine *object)
         }
 
         GtkWidget *scroll = gtk_scrolled_window_new(NULL, NULL);
+        gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(scroll), GTK_SHADOW_OUT);
         gtk_widget_show(scroll);
         gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW (scroll),
                                        GTK_POLICY_NEVER,
                                        GTK_POLICY_AUTOMATIC);
 
+        gtk_container_set_border_width(GTK_CONTAINER(object->list_compl), 2);
         gtk_container_add(GTK_CONTAINER (scroll), object->list_compl);
 
         object->list_compl_items_where = 0;
 
         gtk_widget_show(object->list_compl);
-        int w = gtk_clist_optimal_column_width(GTK_CLIST(object->list_compl),
-                                               0);
+        int w = gtk_clist_optimal_column_width(GTK_CLIST(object->list_compl), 0);
         gtk_widget_set_usize(scroll, w + 40, 150);
 
         gtk_container_add(GTK_CONTAINER(object->win_compl), scroll);
@@ -667,8 +753,9 @@ complete_line(GtkCompletionLine *object)
         y += GTK_WIDGET(object)->allocation.y +
           GTK_WIDGET(object)->allocation.height;
 
-        gtk_widget_popup(object->win_compl, x, y);
-        // gtk_widget_show(object->win_compl);
+        // gtk_widget_popup(object->win_compl, x, y);
+        gtk_window_move(GTK_WINDOW(object->win_compl), x, y);
+        gtk_widget_show(object->win_compl);
 
         gtk_clist_select_row(GTK_CLIST(object->list_compl),
                              object->list_compl_items_where, 0);
@@ -681,6 +768,8 @@ complete_line(GtkCompletionLine *object)
     }
     return GEN_NOT_UNIQUE;
   }
+
+  return GEN_COMPLETION_OK;
 }
 
 GtkWidget *
@@ -803,6 +892,7 @@ search_history(GtkCompletionLine* cl, bool avance, bool begin)
   }
 }
 
+/*
 static int
 inverse_search_history(GtkCompletionLine* cl, bool avance, bool begin)
 {
@@ -818,8 +908,9 @@ inverse_search_history(GtkCompletionLine* cl, bool avance, bool begin)
     return -1;
   }
 }
+*/
 
-static int
+static void
 search_off(GtkCompletionLine* cl)
 {
   int pos = gtk_editable_get_position(GTK_EDITABLE(cl));
@@ -829,9 +920,6 @@ search_off(GtkCompletionLine* cl)
   cl->hist->reset_position();
 }
 
-static gboolean
-on_key_press(GtkCompletionLine *cl, GdkEventKey *event, gpointer data)
-{
 #define STOP_PRESS \
   (gtk_signal_emit_stop_by_name(GTK_OBJECT(cl),   "key_press_event"))
 #define MODE_BEG \
@@ -843,6 +931,26 @@ on_key_press(GtkCompletionLine *cl, GdkEventKey *event, gpointer data)
 #define MODE_SRC \
   (cl->hist_search_mode != GCL_SEARCH_OFF)
 
+static gint tab_pressed(GtkCompletionLine* cl)
+{
+  if (MODE_SRC)
+    search_off(cl);
+  complete_line(cl);
+  return FALSE;
+}
+
+static void
+clear_selection(GtkCompletionLine* cl)
+{
+  int pos = gtk_editable_get_position(GTK_EDITABLE(cl));
+  gtk_editable_select_region(GTK_EDITABLE(cl), pos, pos);
+}
+
+static gboolean
+on_key_press(GtkCompletionLine *cl, GdkEventKey *event, gpointer data)
+{
+  static gint tt_id = -1;
+
   switch (event->type) {
 
    case GDK_KEY_PRESS:
@@ -850,11 +958,20 @@ on_key_press(GtkCompletionLine *cl, GdkEventKey *event, gpointer data)
 
     switch (event->keyval) {
 
+     case GDK_Control_R:
+     case GDK_Control_L:
+     case GDK_Shift_R:
+     case GDK_Shift_L:
+     case GDK_Alt_R:
+     case GDK_Alt_L:
+      break;
+
      case GDK_Tab:
-      if (MODE_SRC) {
-        search_off(cl);
+      if (tt_id != -1) {
+        gtk_timeout_remove(tt_id);
+        tt_id = -1;
       }
-      complete_line(cl);
+      tab_pressed(cl);
       STOP_PRESS;
       return TRUE;
 
@@ -878,10 +995,10 @@ on_key_press(GtkCompletionLine *cl, GdkEventKey *event, gpointer data)
 
      case GDK_space:
      {
+       cl->first_key = 0;
        bool search = MODE_SRC;
-       if (search) {
+       if (search)
          search_off(cl);
-       }
        if (cl->win_compl != NULL) {
          gtk_widget_destroy(cl->win_compl);
          cl->win_compl = NULL;
@@ -891,7 +1008,7 @@ on_key_press(GtkCompletionLine *cl, GdkEventKey *event, gpointer data)
          }
        }
      }
-     return TRUE;
+     return FALSE;
 
      case GDK_Down:
       if (cl->win_compl != NULL) {
@@ -926,10 +1043,11 @@ on_key_press(GtkCompletionLine *cl, GdkEventKey *event, gpointer data)
 
      case GDK_exclam:
       if (!MODE_BEG) {
+        if (!MODE_SRC)
+          gtk_editable_delete_selection(GTK_EDITABLE(cl));
         const char *tmp = gtk_entry_get_text(GTK_ENTRY(cl));
-        if (*tmp != '\0') {
+        if (!(*tmp == '\0' || cl->first_key))
           goto ordinary;
-        }
         cl->hist_search_mode = GCL_SEARCH_BEG;
         cl->hist_word->clear();
         gtk_signal_emit_by_name(GTK_OBJECT(cl), "search_mode");
@@ -978,46 +1096,45 @@ on_key_press(GtkCompletionLine *cl, GdkEventKey *event, gpointer data)
       }
       return FALSE;
 
-     case GDK_Escape:
+     case GDK_Home:
      case GDK_End:
-     {
-       int pos = gtk_editable_get_position(GTK_EDITABLE(cl));
-       gtk_entry_select_region(GTK_ENTRY(cl), pos, pos);
-       if (cl->win_compl != NULL) {
-         gtk_widget_destroy(cl->win_compl);
-         cl->win_compl = NULL;
-         STOP_PRESS;
-         return TRUE;
-       }
-       if (MODE_SRC) {
-         search_off(cl);
-         STOP_PRESS;
-         return TRUE;
-       }
-       return FALSE;
-     }
+      clear_selection(cl);
+      goto ordinary;
+
+     case GDK_Escape:
+      if (MODE_SRC) {
+        search_off(cl);
+      } else if (cl->win_compl != NULL) {
+        gtk_widget_destroy(cl->win_compl);
+        cl->win_compl = NULL;
+      } else {
+        // user cancelled
+        gtk_signal_emit_by_name(GTK_OBJECT(cl), "cancel");
+      }
+      STOP_PRESS;
+      return TRUE;
 
      case GDK_G:
      case GDK_g:
-     case GDK_A:
-     case GDK_a:
-     case GDK_E:
-     case GDK_e:
-      if ((event->state & GDK_CONTROL_MASK) && MODE_SRC) {
+      if (event->state & GDK_CONTROL_MASK) {
         search_off(cl);
-        return FALSE;
+        if (MODE_SRC)
+          STOP_PRESS;
+        return TRUE;
       } else goto ordinary;
 
-     case GDK_Control_R:
-     case GDK_Control_L:
-     case GDK_Shift_R:
-     case GDK_Shift_L:
-     case GDK_Alt_R:
-     case GDK_Alt_L:
-      break;
+     case GDK_E:
+     case GDK_e:
+      if (event->state & GDK_CONTROL_MASK) {
+        search_off(cl);
+        if (MODE_SRC)
+          clear_selection(cl);
+      }
+      goto ordinary;
 
      ordinary:
      default:
+      cl->first_key = 0;
       if (cl->win_compl != NULL) {
         gtk_widget_destroy(cl->win_compl);
         cl->win_compl = NULL;
@@ -1026,14 +1143,21 @@ on_key_press(GtkCompletionLine *cl, GdkEventKey *event, gpointer data)
       if (MODE_SRC) {
         if (event->length > 0) {
           *cl->hist_word += event->string;
-          if (search_history(cl, false, MODE_BEG) <= 0) {
+          if (search_history(cl, false, MODE_BEG) <= 0)
             cl->hist_word->erase(cl->hist_word->length() - 1);
-          }
           STOP_PRESS;
           return TRUE;
-        } else {
+        } else
           search_off(cl);
+      }
+      if (cl->tabtimeout != 0) {
+        if (tt_id != -1) {
+          gtk_timeout_remove(tt_id);
+          tt_id = -1;
         }
+        if (::isprint(*event->string))
+          tt_id = gtk_timeout_add(cl->tabtimeout,
+                                  GtkFunction(tab_pressed), cl);
       }
       break;
     }
@@ -1043,10 +1167,15 @@ on_key_press(GtkCompletionLine *cl, GdkEventKey *event, gpointer data)
     break;
   }
   return FALSE;
+}
 
 #undef STOP_PRESS
 #undef MODE_BEG
 #undef MODE_REW
 #undef MODE_FWD
 #undef MODE_SRC
-}
+
+// Local Variables: ***
+// mode: c++ ***
+// c-basic-offset: 2 ***
+// End: ***
