@@ -1,5 +1,5 @@
 /*****************************************************************************
- *  $Id: main.cc,v 1.24.2.1 2003/06/21 23:14:34 mishoo Exp $
+ *  $Id: main.cc,v 1.24.2.2 2003/08/03 09:53:19 mishoo Exp $
  *  Copyright (C) 2000, Mishoo
  *  Author: Mihai Bazon                  Email: mishoo@fenrir.infoiasi.ro
  *
@@ -168,6 +168,7 @@ run_the_command(const std::string& command, struct gigi* g)
     CT_ESCAPE
   };
   TYPE_CONTEXT context = CT_NORMAL;
+  TYPE_CONTEXT save_context = CT_NORMAL;
   string tmp;
   char c;
 
@@ -180,6 +181,7 @@ run_the_command(const std::string& command, struct gigi* g)
     switch (c) {
      case '\\':
       if (context != CT_ESCAPE) {
+        save_context = context;
         context = CT_ESCAPE;
       } else {
         goto ordinary;
@@ -221,7 +223,7 @@ run_the_command(const std::string& command, struct gigi* g)
      ordinary:
      default:
       if (context == CT_ESCAPE) {
-        context = CT_NORMAL;
+        context = save_context;
         tmp += c;
       } else if (context == CT_QUOTE) {
         tmp += c;
@@ -397,16 +399,23 @@ url_check(GtkCompletionLine *cl, struct gigi *g)
     string url_handler;
 
     if (configuration.get_string(string("URL_") + url_type, url_handler)) {
-      string::size_type j;
+      string::size_type j = 0;
+
+      do {
+        j = url_handler.find("%s", j);
+        if (j != string::npos) {
+          url_handler.replace(j, 2, url);
+        }
+      } while (j != string::npos);
+
+      j = 0;
+      do {
+        j = url_handler.find("%u", j);
+        if (j != string::npos) {
+          url_handler.replace(j, 2, text);
+        }
+      } while (j != string::npos);
 
-      j = url_handler.find("%s");
-      if (j != string::npos) {
-        url_handler.replace(j, 2, url);
-      }
-      j = url_handler.find("%u");
-      if (j != string::npos) {
-        url_handler.replace(j, 2, text);
-      }
       cl->hist->append(text.c_str());
       cl->hist->sync_the_file();
       run_the_command(url_handler.c_str(), g);
